# combo - content management system
# Copyright (C) 2015  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import shutil
import os
import zipfile

from django.conf import settings
from django.contrib import messages
from django.core.files.storage import default_storage
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.template import RequestContext
from django.utils.translation import ugettext as _
from django.views.generic import TemplateView, UpdateView

import ckeditor

from combo.data.models import CellBase, LinkCell, Page
from .models import MomoIconCell, MomoOptions


class MomoManagerView(TemplateView):
    template_name = 'momo/manager_home.html'


def get_page_dict(request, page, manifest):
    cells = CellBase.get_cells(page_id=page.id)

    page_dict = {
        'title': page.title,
        'id': 'page-%s' % page.id,
    }

    link_cells = [x for x in cells if isinstance(x, LinkCell)]
    icon_cells = [x for x in cells if isinstance(x, MomoIconCell)]
    cells = [x for x in cells if not isinstance(x, LinkCell)]

    if cells:
        context = RequestContext(request, {
            'synchronous': True,
            'page': page,
            'page_cells': cells,
            'request': request,
            'site_base': request.build_absolute_uri('/')[:-1],
        })
        page_dict['content'] = '\n'.join([cell.render(context) for cell in cells])

    if link_cells:
        page_dict['seealso'] = []
        for cell in link_cells:
            if cell.link_page:
                # internal link
                page_dict['seealso'].append('page-%s' % cell.link_page.id)
            else:
                # external link
                page_dict['seealso'].append('seealso-%s' % cell.id)
                manifest['_pages'].append({
                    'title': cell.title,
                    'external': True,
                    'url': cell.url,
                    'id': 'seealso-%s' % cell.id})

    if icon_cells:
        page_dict['icon'] = icon_cells[0].icon

    if page.redirect_url:
        page_dict['external'] = True
        page_dict['url'] = page.redirect_url

    if hasattr(page, '_children'):
        children = page._children
    else:
        children = page.get_children()
    if children:
        page_dict['pages'] = []
        for child in children:
            page_dict['pages'].append(get_page_dict(request, child, manifest))

    return page_dict


def generate(request, **kwargs):
    if not default_storage.exists('assets-base.zip'):
        messages.error(request, _('Missing base assets file'))
        return HttpResponseRedirect(reverse('momo-manager-homepage'))

    manifest = {
        'menu': [],
        '_pages': []
    }
    level0_pages = Page.objects.filter(parent=None)

    # the application hierarchy is structured that way:
    # - the home screen is the homepage
    # - the application pages are created from the homepage siblings and their
    #   children
    # - the application menu is created from direct children of the homepage
    children = []
    homepage = None
    for page in level0_pages:
        if page.slug == 'index':
            homepage = page
        else:
            children.append(page)

    if not homepage:
        messages.error(request, _('The homepage needs to be created first.'))
        return HttpResponseRedirect(reverse('momo-manager-homepage'))

    homepage._children = children
    manifest.update(get_page_dict(request, homepage, manifest))

    # get real homepage children to construct the application menu
    menu_children = homepage.get_children()
    for menu_child in menu_children:
        menu_id = 'menu-%s-%s' % (menu_child.id, menu_child.slug)
        manifest['menu'].append(menu_id)
        manifest['_pages'].append({
            'title': menu_child.title,
            'external': True,
            'url': menu_child.redirect_url,
            'id': menu_id,
            })

    options = MomoOptions.get_object()
    manifest['meta'] = {
            'title': options.title or homepage.title,
            'icon': 'icon.png',
            'contact': options.contact_email or 'info@entrouvert.com',
            'updateFreq': options.update_freq or 86400,
            'manifestUrl': request.build_absolute_uri(default_storage.url('index.json')),
            'assetsUrl': request.build_absolute_uri(default_storage.url('assets.zip')),
    }

    if options.icons_on_homepage:
        manifest['display'] = 'icons'

    current_manifest = None
    if default_storage.exists('index.json'):
        with default_storage.open('index.json', mode='r') as fp:
            current_manifest = fp.read()

    new_manifest = json.dumps(manifest, indent=2)
    if new_manifest != current_manifest:
        with default_storage.open('index.json', mode='w') as fp:
            fp.write(new_manifest)
    else:
        messages.info(request, _('No changes were detected.'))
        return HttpResponseRedirect(reverse('momo-manager-homepage'))

    # assets.zip
    if default_storage.exists('assets.zip'):
        zf = zipfile.ZipFile(default_storage.open('assets.zip'))
        existing_files = set([x for x in zf.namelist() if x[0] != '/' and x[-1] != '/'])
        zf.close()
    else:
        existing_files = set([])

    ckeditor_filenames = set(ckeditor.views.get_image_files())
    media_ckeditor_filenames = set(['media/' + x for x in ckeditor_filenames])

    if not media_ckeditor_filenames.issubset(existing_files) or (
            default_storage.modified_time('assets-base.zip') > default_storage.modified_time('assets.zip')):
        # if there are new files, or if the base assets file changed, we
        # generate a new assets.zip
        shutil.copy(default_storage.path('assets-base.zip'),
                default_storage.path('assets.zip.tmp'))
        zf = zipfile.ZipFile(default_storage.path('assets.zip.tmp'), 'a')
        for filename in ckeditor_filenames:
            zf.write(default_storage.path(filename), 'media/' + filename)
        zf.close()
        os.unlink(default_storage.path('assets.zip'))
        os.rename(default_storage.path('assets.zip.tmp'), default_storage.path('assets.zip'))

        messages.info(request, _('A new update (including new assets) has been generated.'))
    else:
        messages.info(request, _('A new update has been generated.'))

    return HttpResponseRedirect(reverse('momo-manager-homepage'))


class OptionsUpdateView(UpdateView):
    model = MomoOptions

    def get_object(self, *args, **kwargs):
        return MomoOptions.get_object()

    def get_success_url(self):
        return reverse('momo-manager-homepage')
